from datetime import datetime

import pkg_resources
pkg_resources.require("SQLAlchemy>=0.4.0")

from turbogears.database import mapper, metadata, session
# import some basic SQLAlchemy classes for declaring the data model
# (see http://www.sqlalchemy.org/docs/04/ormtutorial.html)
from sqlalchemy import Table, Column, ForeignKey
from sqlalchemy.orm import relation
# import some datatypes for table columns from SQLAlchemy
# (see http://www.sqlalchemy.org/docs/04/types.html for more)
from sqlalchemy import String, Unicode, Integer, DateTime
from turbogears import identity


# your data tables
paste_table = Table('paste', metadata,
    Column('guid', String(32), primary_key=True),
    Column('title', Unicode(100)),
    Column('code', Unicode(10000)),
    Column('format_id', ForeignKey('format.name')),
    Column('creator', String(32)),
    Column('created', DateTime, nullable=False, default=datetime.now),
    Column('last_access', DateTime, nullable=False, default=datetime.now)
)

format_table = Table('format', metadata,
    Column('name', String(20), primary_key=True),
    Column('display_name', Unicode(20))
)

style_table = Table('style', metadata,
    Column('name', String(20), primary_key=True),
    Column('display_name', Unicode(20))
)

# your model classes
class Paste(object):
    @classmethod
    def by_guid(self, guid):
        """Look up paste instance by given GUID."""
        return cls.query.get(guid)

    def __repr__(self):
        return "<Paste guid=%r title=%r>" % (self.guid, self.title)

class Format(object):
    @classmethod
    def options(cls):
        """Return all Format instances as an iterator yielding 2-item tuples."""
        return ((f.name, f.display_name)
            for f in cls.query().order_by('display_name'))

    def __str__(self):
        return "%s (%s)" % (self.display_name, self.name)

class Style(object):
    @classmethod
    def options(cls):
        """Return all Style instances as an iterator yielding 2-item tuples."""
        return ((s.name, s.display_name)
            for s in cls.query().order_by('display_name'))

    def __str__(self):
        return "%s (%s)" % (self.display_name, self.name)

# set up mappers between your data tables and classes
mapper(Paste, paste_table,
    properties=dict(format=relation(Format, backref='pastes')))
mapper(Format, format_table)
mapper(Style, style_table)

# the identity schema

visits_table = Table('visit', metadata,
    Column('visit_key', String(40), primary_key=True),
    Column('created', DateTime, nullable=False, default=datetime.now),
    Column('expiry', DateTime)
)

visit_identity_table = Table('visit_identity', metadata,
    Column('visit_key', String(40), primary_key=True),
    Column('user_id', Integer, ForeignKey('tg_user.user_id'), index=True)
)

groups_table = Table('tg_group', metadata,
    Column('group_id', Integer, primary_key=True),
    Column('group_name', Unicode(16), unique=True),
    Column('display_name', Unicode(255)),
    Column('created', DateTime, default=datetime.now)
)

users_table = Table('tg_user', metadata,
    Column('user_id', Integer, primary_key=True),
    Column('user_name', Unicode(16), unique=True),
    Column('email_address', Unicode(255), unique=True),
    Column('display_name', Unicode(255)),
    Column('password', Unicode(40)),
    Column('created', DateTime, default=datetime.now)
)

permissions_table = Table('permission', metadata,
    Column('permission_id', Integer, primary_key=True),
    Column('permission_name', Unicode(16), unique=True),
    Column('description', Unicode(255))
)

user_group_table = Table('user_group', metadata,
    Column('user_id', Integer, ForeignKey('tg_user.user_id',
        onupdate='CASCADE', ondelete='CASCADE'), primary_key=True),
    Column('group_id', Integer, ForeignKey('tg_group.group_id',
        onupdate='CASCADE', ondelete='CASCADE'), primary_key=True)
)

group_permission_table = Table('group_permission', metadata,
    Column('group_id', Integer, ForeignKey('tg_group.group_id',
        onupdate='CASCADE', ondelete='CASCADE'), primary_key=True),
    Column('permission_id', Integer, ForeignKey('permission.permission_id',
        onupdate='CASCADE', ondelete='CASCADE'), primary_key=True)
)


# the identity model

class Visit(object):
    """A visit to your site."""

    @classmethod
    def lookup_visit(cls, visit_key):
        """Look up Visit by given visit key."""
        return cls.query.get(visit_key)


class VisitIdentity(object):
    """A Visit that is linked to a User object."""

    @classmethod
    def by_visit_key(cls, visit_key):
        """Look up VisitIdentity by given visit key."""
        return cls.query.get(visit_key)


class Group(object):
    """An ultra-simple group definition."""

    @classmethod
    def by_group_name(cls, group_name):
        """Look up Group by given group name."""
        return cls.query.filter_by(group_name=group_name).first()
    by_name = by_group_name


class User(object):
    """Reasonably basic User definition.

    Probably would want additional attributes.

    """
    @property
    def permissions(self):
        """Return all permissions os all groups the user belongs to."""
        p = set()
        for g in self.groups:
            p |= set(g.permissions)
        return p

    @classmethod
    def by_email_address(cls, email_address):
        """Look up User by given email address.

        This class method that can be used to search users based on their email
        addresses since it is unique.

        """
        return cls.query.filter_by(email_address=email_address).first()

    @classmethod
    def by_user_name(cls, user_name):
        """Look up User by given user name.

        This class method that permits to search users based on their
        user_name attribute.

        """
        return cls.query.filter_by(user_name=user_name).first()
    by_name = by_user_name

    def _set_password(self, password):
        """Run cleartext password through the hash algorithm before saving."""
        self._password = identity.encrypt_password(password)

    def _get_password(self):
        """Returns password."""
        return self._password

    password = property(_get_password, _set_password)


class Permission(object):
    """A relationship that determines what each Group can do."""

    @classmethod
    def by_permission_name(cls, permission_name):
        """Look up Permission by given permission name."""
        return cls.query.filter_by(permission_name=permission_name).first()
    by_name = by_permission_name


# set up mappers between identity tables and classes

mapper(Visit, visits_table)

mapper(VisitIdentity, visit_identity_table,
        properties=dict(users=relation(User, backref='visit_identity')))

mapper(User, users_table,
        properties=dict(_password=users_table.c.password))

mapper(Group, groups_table,
        properties=dict(users=relation(User,
                secondary=user_group_table, backref='groups')))

mapper(Permission, permissions_table,
        properties=dict(groups=relation(Group,
                secondary=group_permission_table, backref='permissions')))



"""Miscellaneous utility functions."""

__all__ = [
    'absolute_url',
    'add_global_tmpl_vars',
    'code2html',
    'get_user_id',
    'get_server_name',
    'get_static_content',
    'message_of_the_day',
    'txt2html',
    'wrap_text',
]

import logging
import random
import textwrap

from os.path import join

import cherrypy
from turbogears import config, identity, url
from turbogears.release import version as tg_version
from pygments import highlight
from pygments.lexers import get_lexer_by_name
from pygments.styles import get_style_by_name
from pygments.formatters import HtmlFormatter

try:
    from spammcan.rest import HTML
    config.update({'has_docutils': True})
except ImportError:
    config.update({'has_docutils': False})


log = logging.getLogger('spammcan.controllers')
_static_dir = config.get('static_filter.dir', path='/static')


def absolute_url(tgpath='/', params=None, **kw):
    """Returns absolute URL (including schema and host to this server).

    Tries to account for 'Host' header and reverse proxing ('X-Forwarded-Host').

    """

    h = cherrypy.request.headers
    use_xfh = config.get('base_url_filter.use_x_forwarded_host', False)
    if h.get('X-Use-SSL'):
        scheme = 'https'
    else:
        scheme = cherrypy.request.scheme
    base_url = '%s://%s' % (scheme, get_server_name())
    if config.get('base_url_filter.on', False) and not use_xfh:
        base_url = config.get('base_url_filter.base_url').rstrip('/')
    return '%s%s' % (base_url, url(tgpath, params, **kw))

def add_global_tmpl_vars(vars):
    """Add custom global template variables.

    Adds the following variables:

    ``tg.motd`` - The message of the day as a plain string

    """
    vars['motd'] = message_of_the_day()
    vars['abs_url'] = absolute_url
    vars['tg_version'] = tg_version

def code2html(code, format, style=None, cssclass="source", hl_lines=None,
        linenos=False):
    """Syntax highlight given code with Pygments and format as HTML."""
    formatter_options = dict(
        style=style or 'default',
        linenos=linenos,
        cssclass=cssclass,
        lineanchors='L')
    if hl_lines:
        formatter_options['hl_lines'] = hl_lines
    lexer = get_lexer_by_name(format)
    formatter = HtmlFormatter(**formatter_options)
    return highlight(code, lexer, formatter), formatter.get_style_defs(
        '.' + cssclass)

def get_user_id():
    """Return id of current visitor.

    This will be a string with 32-char MD5 hash, which will be retrieved either
    from a request cookie or default to the visit_key of the current identity.

    As a side affect this function sets a response cookie for the user ID
    called 'spammcan_uid' with an expiry time set by the
    'spammcan.uid_cookie_expiry' config setting (defaults to 90 days).

    """
    cookie = cherrypy.request.simple_cookie.get('spammcan_uid')
    log.debug("User ID cookie: %r", cookie)
    if cookie:
        uid = cookie.value
    else:
        uid = identity.current.visit_key
    cherrypy.response.simple_cookie['spammcan_uid'] = uid
    cherrypy.response.simple_cookie['spammcan_uid']['expires'] = config.get(
        'spammcan.uid_cookie_expiry', 3600*24*90)
    return uid

def get_server_name():
    """Return name of the server this application runs on.

    Tries to account for 'Host' header and reverse proxing.

    """
    h = cherrypy.request.headers
    return config.get('server.domain', h.get('X-Forwarded-Host', h['Host']))

def get_static_content(name):
    """Read text file from static directory and convert it to HTML."""
    filename = join(_static_dir, name)
    try:
        fo = open(filename, 'rb')
    except (OSError, IOError):
        text = _(u"Could not read static content resource '%s'.") % name
    else:
        text = fo.read().decode('utf-8')
        fo.close()
    return txt2html(text)

def message_of_the_day(filename='tipoftheday.txt'):
    """Return random message of the day from a text file with one msg per line.
    """
    totd = join(_static_dir, filename)
    return random.choice(filter(None, open(totd).readlines()))

def txt2html(text, use_docutils=True):
    """Try to convert text into HTML with docutils.

    If conversion fails or using docutils is turned off by the configuration,
    return text wrapped in a PRE element with CSS class "plaintext".

    """
    use_docutils = use_docutils and config.get('has_docutils', False)
    if use_docutils:
        try:
            text = HTML(text, report_level=0, initial_header_level=2)
        except Exception, exc:
            log.debug("txt2html failed: %r,", exc)
            use_docutils = False
    return use_docutils and text or '<pre class="plaintext">%s</pre>' % text

def _wrap_line(s, width=100):
    """Wrap single paragraph of text with custom options."""
    return textwrap.wrap(s, width=width, subsequent_indent='  ',
        expand_tabs=True, replace_whitespace=False, break_long_words=False)

def wrap_text(s, width=100):
    """Wrap each paragraph in s with wrap().

    Returns 2-item tuple with a list of line nummers of continuation lines, and
    the wrapped text.

    """
    i = 1; wrapped_lines = []; output= []
    for line in s.splitlines():
        newlines = _wrap_line(line, width) or ['']
        if len(newlines) > 1:
            wrapped_lines.extend(range(i+1,i+len(newlines)))
        i += len(newlines)
        output.extend(newlines)
    return wrapped_lines, "\n".join(output)



#!/home/chris/lib/tg11svn25/bin/python
# -*- coding: utf-8 -*-
"""Start script for the SpammCan TurboGears project.

This script is only needed during development for running from the project
directory. When the project is installed, easy_install will create a
proper start script.
"""

import sys
from spammcan.commands import start, ConfigurationError

if __name__ == "__main__":
    try:
        start()
    except ConfigurationError, exc:
        sys.stderr.write(str(exc))
        sys.exit(1)



# standard library imports
import datetime
import logging
import uuid

from urllib import quote

# third-party imports
from cherrypy import request, NotFound
from turbogears import (config, controllers, expose, error_handler, flash,
    identity, redirect, url, validate, validators, visit)
from turbogears.view import variable_providers

# project specific imports
from spammcan.model import Format, Paste, Style, session
# from spammcan import json
from spammcan.forms import (paste_form, Range, style_select, ValidPasteGuid,
    ValidStyle)
from spammcan.util import (absolute_url, add_global_tmpl_vars, code2html,
    get_user_id, get_server_name, get_static_content, txt2html, wrap_text)


log = logging.getLogger("spammcan.controllers")
variable_providers.append(add_global_tmpl_vars)


class Root(controllers.RootController):
    """The root controller of the application."""

    @expose(template="spammcan.templates.paste")
    @validate(validators=dict(
        paste=ValidPasteGuid(),
        hl=Range(if_invalid=None),
        ln=validators.StringBoolean(if_invalid=True),
        st=ValidStyle(if_invalid=None, convert_value=False),
        wrap=validators.Int(if_invalid=None)))
    def index(self, paste=None, hl=None, ln=True, st=None, wrap=80,
            tg_errors=None):
        """"Show the paste given by GUID."""
        if tg_errors:
            flash(_(u'Paste not found.'))
            paste = None
        #log.debug("Paste: %s", paste)
        if paste is None:
            redirect('/new')

        try:
            if wrap:
                wrapped_lines, code = wrap_text(paste.code, wrap)
                if wrapped_lines:
                    hl = set(hl or [])
                    hl.update(wrapped_lines)
                    hl = list(hl)
                    flash(_(u"Some long lines have been wrapped at %i chars "
                        "and highlighted. Add '?wrap=no' to the paste URL to "
                        "disable line wrapping.") % wrap)
            else:
                code = paste.code
            html, css = code2html(code, paste.format.name, style=st,
                hl_lines=hl, linenos=ln)
        except:
            log.exception('Syntax highlighting error.')
            flash(_(u'Internal error. Syntax highlighting disabled.'))
            html = '<pre class="source">%s</pre>' % paste.code
            css = None
        #log.debug("Pygments output: %s", html)

        # make URL for "Send link" link
        subject = _(u"Link for paste '%s' on %s") % (paste.title or
            _(u"untitled"), get_server_name())
        paste_url = absolute_url(['/paste', paste.guid], st=st)
        mailto_url = 'mailto:?subject=%s&body=%s' % (quote(subject, ''),
            quote(paste_url, ''))

        styles = list(Style.options())
        pagetitle = _(u"Serving paste '%s' fresh from the SpammCan") % (
            paste.title or 'untitled',)
        paste.last_access = datetime.datetime.now()
        session.flush()
        return dict(
            paste = paste,
            pretty_code = html,
            css = css,
            pagetitle = pagetitle,
            mailto_url = mailto_url,
            lineno = not ln,
            style_select = style_select,
            form_values = dict(st=st or 'default'),
            form_params = dict(
                action=url(['/paste', paste.guid]),
                child_args=dict(st={'options': styles})))
    paste = index

    @expose(content_type="text/plain")
    @validate(validators=dict(paste=ValidPasteGuid()))
    def download(self, paste, tg_errors=None, *args, **kw):
        """Serve the paste given by GUID as plain text."""
        if tg_errors:
            raise NotFound
        paste.last_access = datetime.datetime.now()
        session.flush()
        return paste.code

    @expose(template="spammcan.templates.new")
    @validate(validators=dict(paste=ValidPasteGuid(if_invalid=None)))
    def new(self, paste=None, tg_errors=None):
        """Show the form for a new paste."""
        formats = list(Format.options())
        if paste:
            code = paste.code
            format = paste.format.name
            title = paste.title
            if title and not title.startswith('Re: '):
                title = _(u'Re: ') + title
        else:
            code = None
            format = 'text'
            title = None
        return dict(
            form = paste_form,
            values = {
                'format': format,
                'submit': _(u'Spamm me!'),
                'code': code,
                'title': title},
            params = dict(
                action = url('/create'),
                child_args = dict(format={'options': formats})
            ))

    @expose(template="spammcan.templates.paste")
    @validate(form=paste_form)
    @error_handler(new)
    def create(self, code, format=None, title=None):
        """Save submitted paste and redirect to '/paste' to show it."""
        guid = uuid.uuid4().hex
        #log.debug("GUID: %(guid)r, format: %(format)r, title: %(title)r, "
        #    "Code: %(code)r", locals())
        uid = get_user_id()
        paste = Paste(guid=guid, format=format, code=code, title=title,
            creator=uid)
        session.save(paste)
        session.flush()
        redirect(['/paste', guid])

    @expose(template='spammcan.templates.list')
    def mypastes(self, uid=None):
        uid = uid or get_user_id()
        pastes = Paste.query().filter_by(creator=uid).order_by(
            Paste.last_access.desc())
        pagetitle = _(u'Your pastes')
        return dict(pastes=pastes, pagetitle=pagetitle, uid=uid)

    @expose(template="spammcan.templates.static")
    def about(self, package=None):
        """Display information page about the software."""
        content = get_static_content('about.rst')
        return dict(
            pagetitle = _(u'About'),
            heading = _(u'About this site'),
            content = content
        )

    @expose(template="spammcan.templates.static")
    def help(self, package=None):
        """Display usage information for the software."""
        content = get_static_content('help.rst')
        return dict(
            pagetitle = _(u'Help'),
            heading = _(u'How to use the SpammCan'),
            content = content
        )

    @expose(template="spammcan.templates.login")
    def login(self, forward_url=None, *args, **kw):
        """Show the login form or forward user to previously requested page."""

        if forward_url:
            if isinstance(forward_url, list):
                forward_url = forward_url.pop(0)
            else:
                del request.params['forward_url']

        new_visit = visit.current()
        if new_visit:
            new_visit = new_visit.is_new

        if (not new_visit and not identity.current.anonymous
                and identity.was_login_attempted()
                and not identity.get_identity_errors()):
            redirect(forward_url or '/', kw)

        if identity.was_login_attempted():
            if new_visit:
                msg = _(u"Cannot log in because your browser "
                         "does not support session cookies.")
            else:
                msg = _(u"The credentials you supplied were not correct or "
                         "did not grant access to this resource.")
        elif identity.get_identity_errors():
            msg = _(u"You must provide your credentials before accessing "
                     "this resource.")
        else:
            msg = _(u"Please log in.")
            if not forward_url:
                forward_url = request.headers.get("Referer", "/")

        # we do not set the response status here anymore since it
        # is now handled in the identity exception.
        return dict(logging_in=True, message=msg,
            forward_url=forward_url, previous_url=request.path_info,
            original_parameters=request.params)

    @expose()
    def logout(self):
        """Log out the current identity and redirect to start page."""
        identity.current.logout()
        redirect("/")



"""Unit test cases for testing you application's model classes.

If your project uses a database, you should set up database tests similar to
what you see below.

Be sure to set the ``db_uri`` in the ``test.cfg`` configuration file in the
top-level directory of your project to an appropriate uri for your testing
database. SQLite is a good choice for testing, because you can use an in-memory
database which is very fast and the data in it has to be boot-strapped from
scratch every time, so the tests are independant of any pre-existing data.

You can also set the ``db_uri``directly in this test file but then be sure
to do this before you import your model, e.g.::

    from turbogears import testutil, database
    database.set_db_uri("sqlite:///:memory:")
    from spammcan.model import YourModelClass, User, ...
"""

from turbogears.testutil import DBTest

# import the User class defined in the model so we can use it here
from spammcan.model import User
from turbogears.database import session

class TestUser(DBTest):

    def test_creation(self):
        """Object creation should set the name."""
        obj = User(user_name = u"creosote",
                email_address = u"spam@python.not",
                display_name = u"Mr Creosote",
                password = u"Wafer-thin Mint")
        # mark object as 'to be saved'
        session.save(obj)
        # flush marked obj to db
        session.flush()
        retrieved_user = User.by_email_address(u'spam@python.not')
        assert retrieved_user, \
            'User should have been found by email address'

        assert retrieved_user.user_name == u'creosote', \
            "User name should have been creosote, not '%s'" % retrieved_user.user_name
        assert obj.display_name == u"Mr Creosote"




# -*- coding: utf-8 -*-
"""This module contains functions to be called from console script entry points.
"""

import sys
import optparse

from os import getcwd
from os.path import dirname, exists, join

import pkg_resources
try:
    pkg_resources.require("TurboGears>=1.1b1")
except pkg_resources.DistributionNotFound:
    print """\
This is a TurboGears (http://www.turbogears.org) application. It seems that
you either don't have TurboGears installed or it can not be found.

Please check if your PYTHONPATH is set correctly. To install TurboGears, go to
http://docs.turbogears.org/Install and follow the instructions there. If you
are stuck, visit http://docs.turbogears.org/GettingHelp for support options."""
    sys.exit(1)
try:
    pkg_resources.require("SQLAlchemy>=0.4.0")
except pkg_resources.DistributionNotFound:
    from turbogears.util import missing_dependency_error
    print missing_dependency_error('SQLAlchemy')
    sys.exit(1)

import cherrypy
import turbogears

cherrypy.lowercase_api = True


class ConfigurationError(Exception):
    pass


def _read_config(args):
    """Read deployment configuration file.

    First looks on the command line for a desired config file, if it's not on
    the command line, then looks for 'setup.py' in the parent of the directory
    where this module is located.

    If 'setup.py' is there, assumes that the application is started from
    the project directory and should run in development mode and so loads the
    configuration from a file called 'dev.cfg' in the current directory.

    If 'setup.py' is not there, the project is probably installed and the code
    looks first for a file called 'prod.cfg' in the current directory and, if
    this isn't found either, for a default config file called 'default.cfg'
    packaged in the egg.

    """
    setupdir = dirname(dirname(__file__))
    curdir = getcwd()

    if args:
        configfile = args[0]
    elif exists(join(setupdir, "setup.py")):
        configfile = join(setupdir, "dev.cfg")
    elif exists(join(curdir, "prod.cfg")):
        configfile = join(curdir, "prod.cfg")
    else:
        try:
            configfile = pkg_resources.resource_filename(
              pkg_resources.Requirement.parse("spammcan"),
                "config/default.cfg")
        except pkg_resources.DistributionNotFound:
            raise ConfigurationError("Could not find default configuration.")

    turbogears.update_config(configfile=configfile,
        modulename="spammcan.config")

def bootstrap():
    """Example function for loading bootstrap data into the database

    Adapt this to your needs and also uncomment/add a line in your project's
    'setup.py' file that says::

        'bootstrap-spammcan = spammcan.commands:bootstrap',

    """

    optparser = optparse.OptionParser(
        description="Load bootstrap data into the database.")
    optparser.add_option('-C', '--clean', dest="clean", action="store_true",
        help="Purge all data in the database before loading the bootrap data.")
    optparser.add_option('-u', '--user', dest="user", metavar="USERNAME",
        help="Create a default user USERNAME (prompts for password).")
    options, args = optparser.parse_args()
    _read_config(args)
    from bootstrap import bootstrap_model
    bootstrap_model(options.clean, options.user)

def start():
    """Start the CherryPy application server."""

    _read_config(sys.argv[1:])
    from spammcan.controllers import Root
    turbogears.start_server(Root())



# -*- coding: utf-8 -*-
"""A simple pastbin built on TurboGears and Pygments.

General Information
-------------------

SpammCan is yet another pastbin web application. It distinguishes itself from
its competitors by the following features:

* Has syntax highlighting support for over 100 languages thanks to the use of
  Pygments_.

* Is easy to install thanks to setuptools_.

* Uses large, random GUIDs for paste entries in its URLs instead of sequential
  paste numbers to discourage spammers.

  Also detects and rejects spamming attempts with the help of a SpamBayes_
  filter.

* Is built on TurboGears_ 1.1, Genshi_, and SQLAlchemy_.

* Uses a SQLite database by default, but can use any database system supported
  by SQLAlchemy and TurboGears.


Getting the Code
----------------

You can run your own SpammCan server! For more information, downloads, and
source code, visit the project home page at

    http://chrisarndt.de/projects/SpammCan

You can also install SpammCan via its Python Package Index (aka Cheeseshop_)
entry_::

    $ [sudo] easy_install SpammCan

Finally, if you want the latest development code for SpammCan, you can check it
out from the Subversion repository::

    $ svn co svn://chrisarndt.de/projects/SpammCan/trunk SpammCan

.. _cheeseshop: http://pypi.python.org/
.. _entry: http://pypi.python.org/pypi/SpammCan
.. _pygments: http://pygments.pocoo.org/
.. _turbogears: http://www.turbogears.org/
.. _setuptools: http://pypi.python.org/pypi/setuptools
.. _spambayes: http://spambayes.sourceforge.net/
.. _genshi: http://genshi.edgewall.org/
.. _sqlalchemy: http://sqlalchemy.org/

"""

name = "SpammCan"
version = "0.1b"
date = "$Date: 2008-11-10 03:11:16 +0100 (Mon, 10 Nov 2008) $"

description = __doc__.split('\n')[0]
long_description = "\n".join(__doc__.split('\n')[2:])
author = "Christopher Arndt"
email = "chris@chrisarndt.de"
copyright = "Copyright © 2008 Christopher Arndt"

# if it's open source, you might want to specify these
url = "http://chrisarndt.de/project/%s/" % name.lower()
download_url = "http://cheeseshop.python.org/pypi/%s" % name
license = "MIT, Zope Public License (rest.py)"



"""Form widgets and validators are defined here."""

__all__ = [
    'paste_form',
    'style_select'
]

import tgmochikit
from turbogears import validators
from turbogears.widgets import register_static_directory
from tw.api import JSLink, WidgetsList
from tw.forms import ListForm, TableForm, TextArea, TextField, SingleSelectField

from validators import SpamBayesFilter, ValidFormat, ValidStyle


tgmochikit.init(register_static_directory, version="1.4", packed=False)
mochikit_js = [JSLink(link='/tg_widgets/tgmochikit/%s' % p)
    for p in tgmochikit.get_paths()]
styleselect_js = JSLink(modname='spammcan',
    filename="static/javascript/styleselect.js")

class PasteFormSchema(validators.Schema):
    title = validators.UnicodeString(max=100, strip=True)
    code = validators.UnicodeString(max=10000, strip=True, not_empty=True)
    format = ValidFormat(if_invalid=None)
    chained_validators = [SpamBayesFilter('code')]

class PasteFormFields(WidgetsList):
    code = TextArea(suppress_label=True, attrs=dict(cols=75, rows=15))
    title = TextField(label_text=_(u'Title:'),
        attrs=dict(maxlength=100, size=50), help_text=_(u'(Optional)'))
    format = SingleSelectField(label_text=_(u'Format:'),
        options=[('text', _(u'No highliting'))])

class StyleSelectForm(ListForm):
    class fields(WidgetsList):
        st = SingleSelectField(label_text=_(u'Color scheme:'),
            options=[('default', _(u'Default'))])
    submit_text = _(u'Go')
    javascript = mochikit_js + [styleselect_js]

style_select = StyleSelectForm('styleselect')

paste_form = TableForm(
    fields = PasteFormFields(),
    validator = PasteFormSchema()
)



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
from turbogears.finddata import find_package_data

import os
execfile(os.path.join("spammcan", "release.py"))

packages=find_packages()
package_data = find_package_data(where='spammcan',
    package='spammcan')
if os.path.isdir('locales'):
    packages.append('locales')
    package_data.update(find_package_data(where='locales',
        exclude=('*.po',), only_in_packages=False))

setup(
    name=name,
    version=version,
    # uncomment the following lines if you fill them out in release.py
    description=description,
    long_description=long_description,
    author=author,
    author_email=email,
    url=url,
    download_url=download_url,
    license=license,

    install_requires=[
        "TurboGears >= 1.1b1",
        "WebTest",
        "SQLAlchemy >= 0.4.0",
        "Pygments",
        "tw.forms",
        "docutils >= 0.4"
    ],
    zip_safe=False,
    packages=packages,
    package_data=package_data,
    keywords=[
        # Use keywords if you'll be adding your package to the
        # Python Cheeseshop

        # if this has widgets, uncomment the next line
        # 'turbogears.widgets',

        # if this has a tg-admin command, uncomment the next line
        # 'turbogears.command',

        # if this has identity providers, uncomment the next line
        # 'turbogears.identity.provider',

        # If this is a template plugin, uncomment the next line
        # 'python.templating.engines',

        # If this is a full application, uncomment the next line
        'turbogears.app',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Framework :: TurboGears',
        # if this is an application that you'll distribute through
        # the Cheeseshop, uncomment the next line
        'Framework :: TurboGears :: Applications',

        # if this is a package that includes widgets that you'll distribute
        # through the Cheeseshop, uncomment the next line
        # 'Framework :: TurboGears :: Widgets',
    ],
    test_suite='nose.collector',
    entry_points = {
        'console_scripts': [
            'start-spammcan = spammcan.commands:start',
            'bootstrap-spammcan = spammcan.commands:bootstrap',
        ],
    },
    # Uncomment next line and create a default.cfg file in your project dir
    # if you want to package a default configuration in your egg.
    #data_files = [('config', ['default.cfg'])],
    )



__all__ = [
    'bootstrap_model',
    'create_formats',
    'create_tables',
    'create_default_user'
]

from getpass import getpass

from turbogears.database import get_engine, session
from spammcan.model import metadata, Format, Style, User

# functions for populating the database

def bootstrap_model(clean=False, create_user=False):
    create_tables(clean)
    create_formats()
    create_styles()
    if create_user:
        create_default_user(options.user)

def create_default_user(user_name):
    """Create a default user."""
    try:
        u = User.by_user_name(user_name)
    except:
        u = None
    if u:
        print _(u"User '%s' already exists in database.") % user_name
        return
    while True:
        password = getpass(_(u"Enter password for user '%s': ") % user_name).strip()
        password2 = getpass(_(u"Confirm password: ")).strip()
        if password != password2:
            print _(u"Passwords do not match.")
        else:
            break
    u = User(user_name=user_name, display_name=u"Default User",
        email_address=u"%s@nowhere.xyz" % user_name, password=password)
    session.save(u)
    session.flush()
    print _(u"User '%s' created.") % user_name

def create_formats():
    """Populate the 'format' table."""
    from pygments.lexers import get_all_lexers

    if Format.query().count():
        return
    lexers = sorted(get_all_lexers(), key=lambda k: k[0].lower())
    for name, aliases, extensions, mimetypes in lexers:
        f = Format(name=aliases[0], display_name=unicode(name))
        session.save(f)
    session.flush()
    print _(u"Populated 'format' table.")

def create_styles():
    """Populate the 'format' table."""
    from pygments.styles import get_all_styles

    if Style.query().count():
        return
    styles = sorted(get_all_styles(), key=lambda k: k[0].lower())
    for name in styles:
        s = Style(name=name, display_name=unicode(name.capitalize()))
        session.save(s)
    session.flush()
    print _(u"Populated 'style' table.")

def create_tables(drop_all=False):
    """Create all tables defined in the model in the database.

    Optionally drop existing tables before creating them.

    """
    if drop_all:
        print "Dropping all database tables defined in model."
    metadata.bind = get_engine()
    if drop_all:
        metadata.drop_all()
    metadata.create_all()
    print _(u"All database tables defined in model created.")



"""Basic custom validators."""

__all__ = [
    'Range',
    'SpamBayesFilter',
    'ValidFormat',
    'ValidModelValue',
    'ValidPasteGuid',
    'ValidStyle',
]

import logging
import socket
import xmlrpclib

from sqlalchemy.exceptions import InvalidRequestError
from turbogears import config, validators
from turbojson.jsonify import jsonify

from spammcan import model


log = logging.getLogger('spammcan.controllers')


class ValidModelValue(validators.FancyValidator):
    """Check if given value is a valid key to an existing model object.
    """

    class_ = None
    attr = 'id'
    name = 'value'
    convert_value = True

    messages = {
        'notExistant': u"There is no %(class_)s with the %(name)s '%(value)s'."
    }

    def _to_python(self, value, state):
        model_cls = getattr(model, self.class_ or '', None)
        try:
            if model_cls:
                model_inst = model_cls.query.filter_by(
                    **{self.attr: value}).one()
            else:
                msg = u'Model class %s not found.' % self.class_
                log.debug(msg)
                raise ValueError(msg)
        except (ValueError, InvalidRequestError), exc:
            log.debug('Could not validate %s.%s == %r: %r',
                    self.class_, self.class_, value, exc)
            raise validators.Invalid(self.message('notExistant', state,
                class_=self.class_, name=self.name, value=value), value, state)
        if self.convert_value:
            return model_inst
        else:
            return value

class ValidPasteGuid(ValidModelValue):
    class_ = 'Paste'
    attr = 'guid'
    name = 'GUID'
    convert_value = True

class ValidFormat(ValidModelValue):
    class_ = 'Format'
    attr = 'name'
    name = 'name'
    convert_value = True

class ValidStyle(ValidModelValue):
    class_ = 'Style'
    attr = 'name'
    name = 'name'
    convert_value = True

class Range(validators.FancyValidator):
    messages = {
        'invalid': u"Invalid range."
    }

    def _to_python(self, value, state):
        parts = value.split(':', 1)
        try:
            if len(parts) > 1:
                return range(int(parts[0]), int(parts[1])+1)
            else:
                return [int(parts[0])]
        except ValueError, exc:
            raise validators.Invalid(self.message('invalid', state,
                value=value), value, state)

class SpamBayesFilter(validators.FancyValidator):
    """Check given value dict against SpamBayes xmlrpc server."""

    __unpackargs__ = ['error_field']

    sbrpc_url = 'http://localhost:8001/sbrpc'
    spam_threshold = 0.8
    default_score = 0.5

    messages = {
        'isSpam': u'Your form submission was rejected as spam.'
    }

    def validate_python(self, value, state):
        self.sbrpc_url = config.get('sbfilter.sbrpc_url', self.sbrpc_url)
        sb_proxy = xmlrpclib.ServerProxy(self.sbrpc_url, allow_none=True)
        try:
            score = sb_proxy.score(self.convert_values(value), [], [])
        except (socket.error, xmlrpclib.Error, TypeError), exc:
            log.warning("SpamBayes XmlRPC call to 'score()' on '%s' failed: %r",
                self.sbrpc_url, exc)
            score = self.default_score
        log.debug("SpamBayes score for value with keys %s: %.4f", value.keys(),
            score)
        if score >= self.spam_threshold:
            msg = self.message('isSpam', state, value=value, score=score)
            if self.error_field:
                error_dict = {self.error_field: msg}
            else:
                error_dict = None
            raise validators.Invalid(msg, value, state,
                error_dict=error_dict)

    def convert_values(self, values, state=None):
        if isinstance(values, dict):
            result = dict()
            for key, value in values.iteritems():
                if isinstance(value, dict) or hasattr(value, '__iter__'):
                    result[key] = covert_values(value)
                if type(value) not in xmlrpclib.Marshaller.dispatch.keys():
                    result[key] = jsonify(value)
                else:
                    result[key] = value
        elif hasattr(values, '__iter__'):
            result = []
            for value in values:
                if isinstance(value, dict) or hasattr(values, '__iter__'):
                    result.append(covert_values(value))
                elif type(value) not in xmlrpclib.Marshaller.dispatch.keys():
                    result.append(jsonify(value))
                else:
                    result.append(value)
        else:
            result = value
        return result



import unittest
from turbogears import testutil
from spammcan.controllers import Root
from spammcan.model import User


class TestPages(testutil.TGTest):

    root = Root

    def test_method(self):
        """The index method should return a string called 'now'"""
        response = self.app.get('/')
        assert isinstance(response.raw['now'], str)

    def test_index_title(self):
        """"The index page should have the right title."""
        response = self.app.get('/')
        assert "<title>Welcome to TurboGears</title>" in response.body

    def test_login_title(self):
        """The login page should have the right title."""
        response = self.app.get('/login')
        assert "<title>Login</title>" in response
        assert "Please log in." in response
        assert "session cookies" not in response
        assert "credentials" not in response
        assert "not correct" not in response

    def test_login_errors(self):
        """The login page should display the right errors."""
        login = '/login?user_name=nobody&password=wrong&login=Login'
        response = self.app.get(login)
        assert "<title>Login</title>" in response
        assert "session cookies" in response
        cookie = ', '.join(map(str, response.cookies_set.values()))
        response = self.app.get(login, headers=dict(Cookie=cookie))
        assert "<title>Login</title>" in response
        assert "credentials" in response
        assert "not correct" in response

    def test_login_and_logout(self):
        """Login with correct credentials and then logout."""
        User(user_name = u"scott", password = u"tiger",
            display_name = u"Bruce Scott",
            email_address = u"scott@enterprise.com")
        response = self.app.get('/')
        assert "<title>Welcome to TurboGears</title>" in response
        assert 'href="/login"' in response
        assert 'href="/logout"' not in response
        response = self.app.get('/login')
        assert "<title>Login</title>" in response
        assert 'Please log in.' in response
        cookie = ', '.join(map(str, response.cookies_set.values()))
        login = '/login?user_name=scott&password=tiger&login=Login'
        headers = dict(Cookie=cookie)
        response = self.app.get(login, headers=headers, status=302)
        location = response.headers['Location']
        response = self.app.get(location, headers=headers)
        assert "<title>Welcome to TurboGears</title>" in response
        assert "Welcome Bruce Scott" in response
        assert 'href="/login"' not in response
        assert 'href="/logout"' in response
        response = self.app.get('/', headers=headers)
        assert "<title>Welcome to TurboGears</title>" in response
        assert "Welcome Bruce Scott" in response
        assert 'href="/login"' not in response
        assert 'href="/logout"' in response
        response = self.app.get('/logout', headers=headers, status=302)
        location = response.headers['Location']
        response = self.app.get(location, headers=headers)
        assert "<title>Welcome to TurboGears</title>" in response
        assert 'href="/login"' in response
        assert 'href="/logout"' not in response
